function [tform_now,moving_reg,rmse] = pcregisterpda(moving,fixed,params)
%%
% [tform,movingReg,rmse] = pcregisterpda(moving,fixed,params) returns a rigid transformation 
% that registers a moving point cloud to a fixed point cloud.
% 
% Based on:
% Gabriel Agamennoni et al, "Point Clouds Registration with Probabilistic Data Association"
% 2016 IEEE/RSJ International Conference on Intelligent Robots and Systems (IROS)
% Expectation Maximization based probabilistic point cloud registration
% 
% Best performance of this iterative process requires adjusting properties for your data. 
% Consider downsampling point clouds using pcdownsample
%
% Adapted for ECE 286 Sprint 20 UCSD, Prof. Florian Meyer
% Adapted by: Raghav Subbaraman
tic
%% Initializations
% Trivial Affine Transform
tform_now = affine3d;
% M step parameter

if(params.is2d_flag)
    % rotz dx dy
    x0 = [0; 0; 0;];
else
    % rotx roty rotz dx dy dz
    x0 = [0; 0; 0; 0; 0; 0];
end

params.lb = [-180; -180; -180; -100; -100; -100];
params.ub = [180; 180; 180; 100; 100; 100];

%% E - Step

% K- Nearest Neighbor
% Idx = knnsearch(X,Y) finds the nearest neighbor in X for each query point in Y and returns 
% the indices of the nearest neighbors in Idx, a column vector. Idx has the same number of rows as Y.
[knn_idx, knn_dist] = knnsearch(fixed.Location,...
                                transformPointsForward(tform_now,moving.Location),...
                                'K',params.knn);

% Compute Weighting terms
% TODO: Non-gaussian T-distribution based
% Compute PDFs
knn_dist_mr    = knn_dist - max(knn_dist,[],2);
knn_dist_exp   = exp(-(knn_dist_mr.^2)/2);
knn_dist_exp_n = knn_dist_exp./sum(knn_dist_exp,2);

%% M - Step

% Multi-stage sub-problem optimization
for pt_ind = 1:moving.Count
    get_fix_loc = fixed.Location(knn_idx(pt_ind,:)',:);
    
    problem = struct('x0',x0,...
                     'lb',params.lb,...
                     'ub',params.ub,...
                     'solver','lsqnonlin',...
                     'options',params.solver_options);
%     disp(pt_ind)
%     tic
                     
    if(params.is2d_flag)
     [x0,cur_norm] = optimizer_func2d(problem,moving.Location(pt_ind,:),...
                         get_fix_loc,knn_dist_exp_n(pt_ind,:));
    else
     [x0,cur_norm] = optimizer_func(problem,moving.Location(pt_ind,:),...
                         get_fix_loc,knn_dist_exp_n(pt_ind,:));
    end
%     toc
end

% tform_now.T(1:3,1:3) = rotz(x0(3))*roty(x0(2))*rotx(x0(1));
% tform_now.T(4,1:3)   = x0(4:6)';
tform_now.T(1:3,1:3) = rotz(x0(1));
tform_now.T(4,1:2)   = x0(2:3)';

moving_reg = pointCloud(transformPointsForward(tform_now,moving.Location));
rmse = 0;

end

function [x_out,resnorm] = optimizer_func2d(problem,moving,get_fix_loc,knn_dist_exp_n)
    
    tform_now = affine3d;    
    % Levenberg Marquardt Minimization
    problem.objective = @cost_fun;
%     toc
    [x_out,resnorm] = lsqnonlin(problem);
%     toc
    
    function [vec_out] = cost_fun(x)
        tform_now.T(1:3,1:3) = rotz(x(1));
        tform_now.T(4,1:2) = x(2:3)';
        tformed_moving_locs = transformPointsForward(tform_now,moving);
        dist_matrix = (get_fix_loc-tformed_moving_locs).^2;
        dist_matrix = sum(dist_matrix,2);
        vec_out     = knn_dist_exp_n'.*dist_matrix;
    end


end


function [x_out,resnorm] = optimizer_func(problem,moving,get_fix_loc,knn_dist_exp_n)
    
    tform_now = affine3d;    
    % Levenberg Marquardt Minimization
    problem.objective = @cost_fun;
%     toc
    [x_out,resnorm] = lsqnonlin(problem);
%     toc
    
    function [vec_out] = cost_fun(x)
        tform_now.T(1:3,1:3) = rotz(x(3))*roty(x(2))*rotx(x(1));
        tform_now.T(4,1:3) = x(4:6)';
        tformed_moving_locs = transformPointsForward(tform_now,moving);
        dist_matrix = (get_fix_loc-tformed_moving_locs).^2;
        dist_matrix = sum(dist_matrix,2);
        vec_out     = knn_dist_exp_n'.*dist_matrix;
    end


end
    
    


%% SANDBOX
% Code for LSQONLIN
%     tform_now.T(1:3,1:3) = rotz(x0(3))*roty(x0(2))*rotx(x0(1));
%     tform_now.T(4,1:3) = x0(4:6)';
%     tformed_moving_locs = transformPointsForward(tform_now,moving.Location(pt_ind,:));
%     
%     dist_matrix = (get_fix_loc-tformed_moving_locs).^2;
%     dist_matrix = sum(dist_matrix,2);
%     
%     wdist_matrix = knn_dist_exp_n(pt_ind,:)*dist_matrix;


% Code for GLOBAL OPTIMIZATION - DANGEROUS?
%     % Apply RotoTranslation
%     tformed_moving_locs = transformPointsForward(tform_now,moving.Location);
%     % Get NNs and compute distances again
%     get_fix_loc = fixed.Location(reshape(knn_idx.',[],1),:);
%     get_fix_loc_rsh = permute(reshape(get_fix_loc,params.knn,[],3),[2 3 1]);
%     dist_matrix = (get_fix_loc_rsh-tformed_moving_locs).^2;
%     dist_matrix = squeeze(sum(dist_matrix,2));
%     % Apply weighting terms
%     wdist_matrix = dist_matrix.*knn_dist_exp_n;